/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Copyright (c) Patronsoft Limited. All rights reserved. */

#include "StdAfx.h"
#include "IASPL.h"
#include "IASSQL.h"

CPrintLog PL;//Print Log
/* Authentication Read from DB */
// 20080425 - Add quota sharing parameters
extern BOOL IASAuthReadDB(CIASSQL &IASSQL, char *TABLENAME, char *username, long &timeleft,
				   DWORD &accumbw, DWORD &bwcount, DWORD &bwquota,
				   DWORD &accumbwUL, DWORD &bwcountUL, DWORD &bwquotaUL,
				   long &bu, long &bd, char &acl, DWORD &pagenum, DWORD &macmap, char *loginmac,
				   DWORD &ipmap, DWORD &accumbwSha, DWORD &bwcountSha, DWORD &bwquotaSha, char *iBMsg_glo, char *iBMsg_per);

extern BOOL IASLoginWriteDB(CIASSQL &IASSQL, char *TABLENAME, char *username, char *MACorIP);

extern BOOL IASLogoutWriteDB(CIASSQL &IASSQL, char *TABLENAME, char *LOGTABLENAME, char *username,
					  char *szMAC, DWORD bwusage, DWORD bwusageUL, char *loginmac,
					  DWORD bwusageSha, DWORD accumbw, DWORD accumbwUL, DWORD accumbwSha, char *iBMsg_glo, char *iBMsg_per);

/* The registry value where this extension is registered. */
LPCWSTR pwszDllType = AUTHSRV_EXTENSIONS_VALUE_W;

/* Global handle to the LSA. This is initialized once at start-up and reused
 * until shutdown. */
LSA_HANDLE hPolicy = NULL;

DWORD
WINAPI
RadiusExtensionInit( VOID )
{
	PL.DP("** == RadiusExtensionInit ==");
   NTSTATUS status;
   LSA_OBJECT_ATTRIBUTES objectAttributes;

   memset(&objectAttributes, 0, sizeof(objectAttributes));
   status = LsaOpenPolicy(
               NULL,
               &objectAttributes,
               POLICY_LOOKUP_NAMES,
               &hPolicy
               );

   return LsaNtStatusToWinError(status);
}

VOID
WINAPI
RadiusExtensionTerm( VOID )
{
	PL.DP("** == RadiusExtensionTerm ==");
   LsaClose(hPolicy);
   hPolicy = NULL;
}

BOOL AuthAccept(PRADIUS_EXTENSION_CONTROL_BLOCK pECB)
{
	PL.DP("** == AuthAccept ==");
	PRADIUS_ATTRIBUTE_ARRAY pInAttrs;			//request RADIUS packet from FS
	PRADIUS_ATTRIBUTE_ARRAY pInAttrs_resp;		//IAS response packet to FS
	DWORD attrCount = 0,						//number of attribute in request packet
		attrCount_resp = 0,						//number of attribute in request packet
		VID = 0;								//FS vendor ID
	char ConfigFile[_MAX_FNAME] = "";			//IASConfig.ini file path
	char CONNSTR[_MAX_FNAME] = "FILEDSN=";		//connection string
	char TABLENAME[_MAX_FNAME] = "";			//user table name
	char tmpVal[_MAX_FNAME] = "";				//temp store return value;
	int ret = 0;
	const RADIUS_ATTRIBUTE *pAttr_add, *pAttr;	//temp attribute pointer
	RADIUS_ATTRIBUTE Attr_add;					//Attribute will be added
	RADIUS_VSA_FORMAT *vendor_add;				//vendor specific attribute will be added
	char usrname[41] = "";						//user name
	char tempStr[_MAX_FNAME] = "";				//temp string store
	char NASIdentifier[_MAX_FNAME] = "";		//NAS identifier
	long timel = -1, bu = -1, bd = -1;
	DWORD accumbw = 0,
		bwcount = 0,
		bwquota = 0;
	DWORD accumbwUL = 0,
		bwcountUL = 0,
		bwquotaUL = 0;
	DWORD pagenum = 0;
	char acl = 'N';
	int result = 0;
	CIASSQL isql;
	DWORD accumbwSha = 0, bwcountSha = 0, bwquotaSha = 0;
	
	DWORD ipmap = 0, macmap = 0;
	char logmac[16] = "";
	
	char iBMsg_glo[41] ="";
	char iBMsg_per[41] = "";
	
	int sz = 0;
	
	ZeroMemory( logmac, 16 );
	
	ZeroMemory( iBMsg_glo, 41 );
	ZeroMemory( iBMsg_per, 41 );
	
	ZeroMemory( ConfigFile, _MAX_FNAME );

	// Construct IASConfig.ini path
	sprintf( ConfigFile, "%s\\IASConfig.ini", PL.fname );

	// Obtain Vendor ID from IASConfig.ini
	VID = GetPrivateProfileInt( "radius", "VendorID", 0, ConfigFile );
	if( VID==0 ){
		PL.DP("** AuthAccept: Can not obtain Vendor ID from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}
	//else
	//	PL.DP("** AuthAccept: VENDOR ID[%d]", VID);

	// Obtain File DSN from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "Datasource", "", tmpVal, _MAX_FNAME,
		ConfigFile );
	if( ret>0 ){
		tmpVal[ret] = '\0';
		strcat( CONNSTR, tmpVal );
		strcat( CONNSTR, ";" );
	}else{
		PL.DP("** AuthAccept: Can not obtain File DSN from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}

	// Obtain table name from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "usrtable", "", TABLENAME, _MAX_FNAME,
		ConfigFile );
	if( ret>0 ){
		TABLENAME[ret] = '\0';
	}else{
		PL.DP("** AuthAccept: Can not obtain Table name from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}

	// Obtain Request packet pointer and size
	pInAttrs = pECB->GetRequest(pECB);
	attrCount = pInAttrs->GetSize(pInAttrs);
	// Obtain Response packet pointer and size
	pInAttrs_resp = pECB->GetResponse(pECB,rcAccessAccept);
	attrCount_resp = pInAttrs_resp->GetSize(pInAttrs_resp);

	// Obtain NAS Identifier Attribute value
	pAttr = RadiusFindFirstAttribute( pInAttrs, ratNASIdentifier );
	memcpy( tempStr, pAttr->lpValue,
		_MAX_FNAME>pAttr->cbDataLength?pAttr->cbDataLength:_MAX_FNAME );

	if( _MAX_FNAME>pAttr->cbDataLength )
		tempStr[pAttr->cbDataLength] = '\0';
	else
		tempStr[_MAX_FNAME-1] = '\0';

	// Obtain NAS Identifier from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "NAS_IDENTIFIER", "", NASIdentifier, _MAX_FNAME,
		ConfigFile );
	if( !(ret>0) ){
		strcpy( NASIdentifier, "FirstSpot" );
		PL.DP("** AuthAccept: Can not obtain NAS Identifier from IASConfig.ini[%d]",
			ret);
		PL.DP("** AuthAccept: Use FirstSpot as Default NAS Identifier" );
	}
	NASIdentifier[ret] = '\0';

	if( strcmp( tempStr, NASIdentifier )==0 )
	{
		// Obtain User Name
		ZeroMemory( usrname, 41 );
		pAttr = RadiusFindFirstAttribute( pInAttrs, ratUserName );
		memcpy( usrname, pAttr->lpValue,
			41>pAttr->cbDataLength?pAttr->cbDataLength:41 );
		usrname[40] = '\0';

		isql.connect(CONNSTR);
		if( isql.connected )
		{
			IASAuthReadDB(isql, TABLENAME, usrname, timel, accumbw, bwcount, bwquota,
				accumbwUL, bwcountUL, bwquotaUL, bu, bd, acl, pagenum, macmap, logmac, ipmap, 
				accumbwSha, bwcountSha, bwquotaSha, 
				iBMsg_glo, iBMsg_per);
			isql.disconnect();
		}
		else
		{
			PL.DP( "** AuthAccept: Cannot connect database" );
			return ERROR_INVALID_PARAMETER;
		}

		PL.DP("** AuthAccept: NAS Identifier[%s]", NASIdentifier);
		PL.DP("** AuthAccept: VENDOR ID[%d]", VID);
		PL.DP("** AuthAccept: File DSN[%s]", CONNSTR);
		PL.DP("** AuthAccept: Table name[%s]", TABLENAME);
		PL.DP("** AuthAccept: User Name[%s]", usrname);
		PL.DP("** AuthAccept: ACL[%c]", acl);
		PL.DP("** AuthAccept: pagenum[%u]", pagenum);
		if( timel == -1 )
			PL.DP("** AuthAccept: Time Left Unlimit");
		else
			PL.DP("** AuthAccept: Time Left[%d]mins", timel);

		if( bwcount==0 )
			PL.DP("** AuthAccept: Download Bandwidth Counting Mode Off");
		else if( bwcount==1 )
		{
			PL.DP("** AuthAccept: Download Bandwidth Counting Mode On, Quota Limitation Off");
			if( accumbw>0 )
				PL.DP("** AuthAccept: Download Accumulative Bandwidth[%u]KBytes", accumbw);
		}
		else if( bwcount==2 )
		{
			PL.DP("** AuthAccept: Download Bandwidth Counting Mode On, Quota Limitation On");
			if( accumbw>0 )
				PL.DP("** AuthAccept: Download Accumulative Bandwidth[%u]KBytes", accumbw);
			if( bwquota>0 )
				PL.DP("** AuthAccept: Download Bandwidth Quota[%u]KBytes", bwquota);
		}
		else if( bwcountUL==-1 )
			PL.DP("** AuthAccept: Use Download Global Bandwidth Counting");

		if( bwcountUL==0 )
			PL.DP("** AuthAccept: Upload Bandwidth Counting Mode Off");
		else if( bwcountUL==1 )
		{
			PL.DP("** AuthAccept: Upload Bandwidth Counting Mode On, Quota Limitation Off");
			if( accumbwUL>0 )
				PL.DP("** AuthAccept: Upload Accumulative Bandwidth[%u]KBytes", accumbwUL);
		}
		else if( bwcountUL==2 )
		{
			PL.DP("** AuthAccept: Upload Bandwidth Counting Mode On, Quota Limitation On");
			if( accumbwUL>0 )
				PL.DP("** AuthAccept: Upload Accumulative Bandwidth[%u]KBytes", accumbwUL);
			if( bwquotaUL>0 )
				PL.DP("** AuthAccept: Upload Bandwidth Quota[%u]KBytes", bwquotaUL);
		}
		else if( bwcountUL==-1 )
			PL.DP("** AuthAccept: Upload Use Global Bandwidth Counting");

		// 20080425 - Add quota sharing part
		if (bwcountSha == 0) {
			PL.DP("** AuthAccept: Sharing Bandwidth Counting Mode Off");
		} else if (bwcountSha == 1) {
			PL.DP("** AuthAccept: Sharing Bandwidth Counting Mode On, Quota Limitation Off");
			if (accumbwSha > 0) {
				PL.DP("** AuthAccept: Sharing Accumulative Bandwidth[%u]KBytes", accumbwSha);
			}
		} else if (bwcountSha == 2) {
			PL.DP("** AuthAccept: Sharing Bandwidth Counting Mode On, Quota Limitation On");
			if (accumbwSha > 0) {
				PL.DP("** AuthAccept: Sharing Accumulative Bandwidth[%u]KBytes", accumbwSha);
			}
			if (bwquotaSha > 0) {
				PL.DP("** AuthAccept: Sharing Bandwidth Quota[%u]KBytes", bwquotaSha);
			}
		}

		if( bu == -1 )
			PL.DP("** AuthAccept: Use Global Bandwidth Upload Thottling");
		else if( bu == 0 )
			PL.DP("** AuthAccept: Bandwidth Upload Thottling Unlimit");
		else
			PL.DP("** AuthAccept: Bandwidth Upload Thottling[%d]KBytes", bu);

		if( bd == -1 )
			PL.DP("** AuthAccept: Use Global Bandwidth Download Thottling");
		else if( bd == 0 )
			PL.DP("** AuthAccept: Bandwidth Download Thottling Unlimit");
		else
			PL.DP("** AuthAccept: Bandwidth Download Thottling[%d]KBytes", bd);
			
		if (ipmap == 1)
			PL.DP("** AuthAccept: Enable Real IP Address Mapping ");
		else
			PL.DP("** AuthAccept: Disablele Real IP Address Mapping ");
			
		if (macmap == 1) {
			PL.DP("** AuthAccept: Enable User bind mac address ");
			if (logmac == NULL || strcmp(logmac, "")==0 || strcmp(logmac, "0")==0){
				PL.DP("** AuthAccept: User account will bind to first login MAC address ");
			}else {
				PL.DP("** AuthAccept: User account bind to [%s] mac address ",logmac);
			}
		}	else {
			PL.DP("** AuthAccept: Disable User bind mac address ");
		}
		// Add Attributes
		DWORD VendorID = htonl( VID );//vendor ID from host order to network order

		// Add timeleft attribute
		long timeleft = htonl( timel );
		Attr_add.dwAttrType = ratVendorSpecific;
		Attr_add.fDataType = rdtString;
		Attr_add.lpValue = new char[_MAX_FNAME]; //* size will affect the other 
		vendor_add = (RADIUS_VSA_FORMAT*)Attr_add.lpValue;
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_TIMELEFT;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&timeleft, 4 );
		Attr_add.cbDataLength = 10;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );

		// Add bandwidth upload limit attribute
		long btul = htonl( bu );
		Attr_add.dwAttrType = ratVendorSpecific;
		Attr_add.fDataType = rdtString;
		ZeroMemory( (char*)Attr_add.lpValue, 10 );
		vendor_add = (RADIUS_VSA_FORMAT*)Attr_add.lpValue;
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_ULLIMIT;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&btul, 4 );
		Attr_add.cbDataLength = 10;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );

		// Add bandwidth download limit attribute
		long btdl = htonl( bd );
		Attr_add.dwAttrType = ratVendorSpecific;
		Attr_add.fDataType = rdtString;
		ZeroMemory( (char*)Attr_add.lpValue, 10 );
		vendor_add = (RADIUS_VSA_FORMAT*)Attr_add.lpValue;
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_DLLIMIT;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&btdl, 4 );
		Attr_add.cbDataLength = 10;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );

		if( bwcount==0 )
		{
			// Add bandwidth counting mode attribute
			ZeroMemory( (char*)Attr_add.lpValue, 10 );

			long bwc = htonl( bwcount );
			memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
			vendor_add->VendorType = FS_AUTH_BWCOUNT;
			vendor_add->VendorLength = 6;
			memcpy( vendor_add->AttributeSpecific, (BYTE*)&bwc, 4 );
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
			if ( result==E_ACCESSDENIED )
				PL.DP("E_ACCESSDENIED" );
			else if ( result==ERROR_INVALID_PARAMETER )
				PL.DP("ERROR_INVALID_PARAMETER" );
		}
		else if( bwcount>0 )
		{
			// Add bandwidth counting mode attribute
			ZeroMemory( (char*)Attr_add.lpValue, 10 );

			long bwc = htonl( bwcount );
			memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
			vendor_add->VendorType = FS_AUTH_BWCOUNT;
			vendor_add->VendorLength = 6;
			memcpy( vendor_add->AttributeSpecific, (BYTE*)&bwc, 4 );
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
			if ( result==E_ACCESSDENIED )
				PL.DP("E_ACCESSDENIED" );
			else if ( result==ERROR_INVALID_PARAMETER )
				PL.DP("ERROR_INVALID_PARAMETER" );

			// Add accumulative bandwidth attribute
			ZeroMemory( (char*)Attr_add.lpValue, 10 );

			DWORD abw = htonl( accumbw );
			memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
			vendor_add->VendorType = FS_AUTH_ACCUMBW;
			vendor_add->VendorLength = 6;
			memcpy( vendor_add->AttributeSpecific, (BYTE*)&abw, 4 );
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
			if ( result==E_ACCESSDENIED )
				PL.DP("E_ACCESSDENIED" );
			else if ( result==ERROR_INVALID_PARAMETER )
				PL.DP("ERROR_INVALID_PARAMETER" );

			if( bwcount>1 )
			{
				// Add bandwidth quota attribute
				ZeroMemory( (char*)Attr_add.lpValue, 10 );

				DWORD bwq = htonl( bwquota );
				memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
				vendor_add->VendorType = FS_AUTH_BWQUOTA;
				vendor_add->VendorLength = 6;
				memcpy( vendor_add->AttributeSpecific, (BYTE*)&bwq, 4 );
				Attr_add.cbDataLength = 10;
				pAttr_add = &Attr_add;
				result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
				if ( result==E_ACCESSDENIED )
					PL.DP("E_ACCESSDENIED" );
				else if ( result==ERROR_INVALID_PARAMETER )
					PL.DP("ERROR_INVALID_PARAMETER" );
			}
		}

		if( bwcountUL==0 )
		{
			// Add bandwidth counting mode attribute
			ZeroMemory( (char*)Attr_add.lpValue, 10 );

			long bwcUL = htonl( bwcountUL );
			memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
			vendor_add->VendorType = FS_AUTH_BWCOUNTUL;
			vendor_add->VendorLength = 6;
			memcpy( vendor_add->AttributeSpecific, (BYTE*)&bwcUL, 4 );
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
			if ( result==E_ACCESSDENIED )
				PL.DP("E_ACCESSDENIED" );
			else if ( result==ERROR_INVALID_PARAMETER )
				PL.DP("ERROR_INVALID_PARAMETER" );
		}
		else if( bwcountUL>0 )
		{
			// Add bandwidth counting mode attribute
			ZeroMemory( (char*)Attr_add.lpValue, 10 );

			long bwcUL = htonl( bwcountUL );
			memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
			vendor_add->VendorType = FS_AUTH_BWCOUNTUL;
			vendor_add->VendorLength = 6;
			memcpy( vendor_add->AttributeSpecific, (BYTE*)&bwcUL, 4 );
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
			if ( result==E_ACCESSDENIED )
				PL.DP("E_ACCESSDENIED" );
			else if ( result==ERROR_INVALID_PARAMETER )
				PL.DP("ERROR_INVALID_PARAMETER" );

			// Add accumulative bandwidth attribute
			ZeroMemory( (char*)Attr_add.lpValue, 10 );

			DWORD abwUL = htonl( accumbwUL );
			memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
			vendor_add->VendorType = FS_AUTH_ACCUMBWUL;
			vendor_add->VendorLength = 6;
			memcpy( vendor_add->AttributeSpecific, (BYTE*)&abwUL, 4 );
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
			if ( result==E_ACCESSDENIED )
				PL.DP("E_ACCESSDENIED" );
			else if ( result==ERROR_INVALID_PARAMETER )
				PL.DP("ERROR_INVALID_PARAMETER" );

			if( bwcountUL>1 )
			{
				// Add bandwidth quota attribute
				ZeroMemory( (char*)Attr_add.lpValue, 10 );

				DWORD bwqUL = htonl( bwquotaUL );
				memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
				vendor_add->VendorType = FS_AUTH_BWQUOTAUL;
				vendor_add->VendorLength = 6;
				memcpy( vendor_add->AttributeSpecific, (BYTE*)&bwqUL, 4 );
				Attr_add.cbDataLength = 10;
				pAttr_add = &Attr_add;
				result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
				if ( result==E_ACCESSDENIED )
					PL.DP("E_ACCESSDENIED" );
				else if ( result==ERROR_INVALID_PARAMETER )
					PL.DP("ERROR_INVALID_PARAMETER" );
			}
		}

		// 20080425 - Add quota sharing part
		if (bwcountSha == 0) {
			// Add bandwidth counting mode attribute
			ZeroMemory((char*)Attr_add.lpValue, 10);

			long bwcSha = htonl(bwcountSha);
			memcpy(vendor_add->VendorId, (BYTE*)&VendorID, 4);
			vendor_add->VendorType = FS_AUTH_BWCOUNTSHARE;
			vendor_add->VendorLength = 6;
			memcpy(vendor_add->AttributeSpecific, (BYTE*)&bwcSha, 4);
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add(pInAttrs_resp, pAttr_add);
			if (result == E_ACCESSDENIED) {
				PL.DP("E_ACCESSDENIED");
			} else if (result == ERROR_INVALID_PARAMETER) {
				PL.DP("ERROR_INVALID_PARAMETER");
			}
		} else if (bwcountSha > 0) {
			// Add bandwidth counting mode attribute
			ZeroMemory((char*)Attr_add.lpValue, 10);

			long bwcSha = htonl(bwcountSha);
			memcpy(vendor_add->VendorId, (BYTE*)&VendorID, 4);
			vendor_add->VendorType = FS_AUTH_BWCOUNTSHARE;
			vendor_add->VendorLength = 6;
			memcpy(vendor_add->AttributeSpecific, (BYTE*)&bwcSha, 4);
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add(pInAttrs_resp, pAttr_add);
			if (result == E_ACCESSDENIED) {
				PL.DP("E_ACCESSDENIED");
			} else if (result == ERROR_INVALID_PARAMETER) {
				PL.DP("ERROR_INVALID_PARAMETER");
			}

			// Add accumulative bandwidth attribute
			ZeroMemory((char*)Attr_add.lpValue, 10);

			DWORD abwSha = htonl(accumbwSha);
			memcpy(vendor_add->VendorId, (BYTE*)&VendorID, 4);
			vendor_add->VendorType = FS_AUTH_ACCUMBWSHARE;
			vendor_add->VendorLength = 6;
			memcpy(vendor_add->AttributeSpecific, (BYTE*)&abwSha, 4);
			Attr_add.cbDataLength = 10;
			pAttr_add = &Attr_add;
			result = pInAttrs_resp->Add(pInAttrs_resp, pAttr_add);
			if (result == E_ACCESSDENIED) {
				PL.DP("E_ACCESSDENIED");
			} else if (result == ERROR_INVALID_PARAMETER) {
				PL.DP("ERROR_INVALID_PARAMETER");
			}

			if (bwcountSha > 1) {
				// Add bandwidth quota attribute
				ZeroMemory((char*)Attr_add.lpValue, 10);

				DWORD bwqSha = htonl(bwquotaSha);
				memcpy(vendor_add->VendorId, (BYTE*)&VendorID, 4);
				vendor_add->VendorType = FS_AUTH_BWQUOTASHARE;
				vendor_add->VendorLength = 6;
				memcpy(vendor_add->AttributeSpecific, (BYTE*)&bwqSha, 4);
				Attr_add.cbDataLength = 10;
				pAttr_add = &Attr_add;
				result = pInAttrs_resp->Add(pInAttrs_resp, pAttr_add);
				if (result == E_ACCESSDENIED) {
					PL.DP("E_ACCESSDENIED");
				} else if (result == ERROR_INVALID_PARAMETER) {
					PL.DP("ERROR_INVALID_PARAMETER");
				}
			}
		}

		// Add ACL attribute
		ZeroMemory( (char*)Attr_add.lpValue, 10 );

		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_ACL;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&acl, 1 );
		Attr_add.cbDataLength = 7;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );

		// Add pagenum attribute
		ZeroMemory( (char*)Attr_add.lpValue, 10 );

		pagenum = htonl( pagenum );
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_PRINTQUOTA;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&pagenum, 4 );
		Attr_add.cbDataLength = 10;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );

		// Add Real IP Address Mapping 
		ZeroMemory( (char*)Attr_add.lpValue, 10 );

		ipmap = htonl( ipmap );
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_IPMAP;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&ipmap, 4 );
		Attr_add.cbDataLength = 10;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );

		// User Bind Mac address Mapping 
		ZeroMemory( (char*)Attr_add.lpValue, 10 );

		macmap = htonl( macmap );
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_MACMAP;
		vendor_add->VendorLength = 6;
		memcpy( vendor_add->AttributeSpecific, (BYTE*)&macmap, 4 );
		Attr_add.cbDataLength = 10;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );
		
		// Bind Mac Address
		sz = 0;
		sz = 6 + strlen(logmac) ;
		PL.DP("AuthAccept: Bind Mac Address [%d] -- Length [%s]",sz, logmac );		
		
		ZeroMemory( (char*)Attr_add.lpValue, sz );
		
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_LOGINMAC;
		vendor_add->VendorLength = 6;

		memcpy( vendor_add->AttributeSpecific, (BYTE*)logmac,strlen(logmac)   );
		Attr_add.cbDataLength = sz;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );
			
		// Global in-browser message state
		sz = 0;
		sz = 6 + strlen(iBMsg_glo) ;
		PL.DP("AuthAccept: Global in-browser message state [%d] -- Length [%s]",sz, iBMsg_glo );		
		
		ZeroMemory( (char*)Attr_add.lpValue, sz );
		
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_IBMSG_GLO;
		vendor_add->VendorLength = 6;

		memcpy( vendor_add->AttributeSpecific, (BYTE*)iBMsg_glo,strlen(iBMsg_glo)   );
		Attr_add.cbDataLength = sz;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );
			
		// Personal in-browser message state
		sz = 0;
		sz = 6 + strlen(iBMsg_per) ;
		PL.DP("AuthAccept: Personal in-browser message state [%d] -- Length [%s]",sz, iBMsg_per );		
		
		ZeroMemory( (char*)Attr_add.lpValue, sz );
		
		memcpy( vendor_add->VendorId, (BYTE*)&VendorID, 4 );
		vendor_add->VendorType = FS_AUTH_IBMSG_PER;
		vendor_add->VendorLength = 6;

		memcpy( vendor_add->AttributeSpecific, (BYTE*)iBMsg_per,strlen(iBMsg_per)   );
		Attr_add.cbDataLength = sz;
		pAttr_add = &Attr_add;
		result = pInAttrs_resp->Add( pInAttrs_resp, pAttr_add );
		if ( result==E_ACCESSDENIED )
			PL.DP("E_ACCESSDENIED" );
		else if ( result==ERROR_INVALID_PARAMETER )
			PL.DP("ERROR_INVALID_PARAMETER" );
		
		delete (char*)Attr_add.lpValue;

		PL.DP("** AuthAccept: Attribute(s)[%u]", pInAttrs_resp->GetSize(pInAttrs_resp));
	}
	pAttr = NULL;

	return NO_ERROR;
}

BOOL Accounting( PRADIUS_EXTENSION_CONTROL_BLOCK pECB )
{
	PL.DP("**== Accounting ==");
	PRADIUS_ATTRIBUTE_ARRAY pInAttrs;			//request RADIUS packet from FS
	DWORD attrCount = 0,						//number of attribute in request packet
		attrCount_resp = 0,						//number of attribute in request packet
		VID = 0;								//FS vendor ID
	char ConfigFile[_MAX_FNAME] = "";			//IASConfig.ini file path
	char CONNSTR[_MAX_FNAME] = "FILEDSN=";		//connection string
	char TABLENAME[_MAX_FNAME] = "";			//user table name
	char LOGTABLENAME[_MAX_FNAME] = "";			//user log table name
	char tmpVal[_MAX_FNAME] = "";				//temp store return value;
	int ret = 0;
	const RADIUS_ATTRIBUTE *pAttr;				//temp attribute pointer
	char usrname[41] = "";						//user name
	char tempStr[_MAX_FNAME] = "";				//temp string store
	char NASIdentifier[_MAX_FNAME] = "";		//NAS identifier
	char CallingStationID[16]="";				//Calling Station ID - IP or MAC
	long AcctStatusType = 0;					//Accouting Status Type
	RADIUS_VSA_FORMAT *FSVendor;				//FS specific vendor attribute
	DWORD bwusage = 0, bwusageUL = 0;			//user bandwidth usage
	DWORD printQuota = 0;
	BOOL result = FALSE;
	CIASSQL isql;
	DWORD bwusageSha = 0;						// 20080428

	// 20090422 - Set zero
	DWORD accumbw = 0;						// 20080428
	DWORD accumbwUL = 0;						// 20080428
	DWORD accumbwSha = 0;						// 20080428
	
	char logmac[16] ="";
	
	char iBMsg_glo[41] ="";
	char iBMsg_per[41] = "";
	
	ZeroMemory( iBMsg_glo, 41 );
	ZeroMemory( iBMsg_per, 41 );

	ZeroMemory( ConfigFile, _MAX_FNAME );

	// Construct IASConfig.ini path
	sprintf( ConfigFile, "%s\\IASConfig.ini", PL.fname );

	// Obtain File DSN from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "Datasource", "", tmpVal, _MAX_FNAME,
		ConfigFile );
	if( ret>0 ){
		tmpVal[ret] = '\0';
		strcat( CONNSTR, tmpVal );
		strcat( CONNSTR, ";" );
	}else{
		PL.DP("** Accounting: Can not obtain File DSN from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}

	// Obtain table name from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "usrtable", "", TABLENAME, _MAX_FNAME,
		ConfigFile );
	if( ret>0 ){
		TABLENAME[ret] = '\0';
	}else{
		PL.DP("** Accounting: Can not obtain Table name from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}

	// Obtain log table name from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "usrlogtable", "", LOGTABLENAME, _MAX_FNAME,
		ConfigFile );
	if( ret>0 ){
		LOGTABLENAME[ret] = '\0';
	}else{
		PL.DP("** Accounting: Can not obtain Log Table name from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}

	// Obtain Request packet pointer and size
	pInAttrs = pECB->GetRequest(pECB);
	attrCount = pInAttrs->GetSize(pInAttrs);

	// Obtain NAS Identifier Attribute value
	pAttr = RadiusFindFirstAttribute( pInAttrs, ratNASIdentifier );
	memcpy( tempStr, pAttr->lpValue,
		_MAX_FNAME>pAttr->cbDataLength?pAttr->cbDataLength:_MAX_FNAME );

	if( _MAX_FNAME>pAttr->cbDataLength )
		tempStr[pAttr->cbDataLength] = '\0';
	else
		tempStr[_MAX_FNAME-1] = '\0';

	// Obtain NAS Identifier from IASConfig.ini
	ret = GetPrivateProfileString( "radius", "NAS_IDENTIFIER", "", NASIdentifier, _MAX_FNAME,
		ConfigFile );
	if( !(ret>0) ){
		strcpy( NASIdentifier, "FirstSpot" );
		PL.DP("** Accounting: Can not obtain NAS Identifier from IASConfig.ini[%d]", ret);
		PL.DP("** Accounting: Use FirstSpot as Default NAS Identifier" );
	}
	NASIdentifier[ret] = '\0';

	// Obtain Vendor ID from IASConfig.ini
	VID = GetPrivateProfileInt( "radius", "VendorID", 0, ConfigFile );
	if( VID==0 ){
		PL.DP("** Accounting: Can not obtain Vendor ID from IASConfig.ini" );
		return ERROR_INVALID_PARAMETER;
	}

	if( strcmp( tempStr, NASIdentifier )==0 ){
		// Obtain User Name
		ZeroMemory( usrname, 41 );
		pAttr = RadiusFindFirstAttribute( pInAttrs, ratUserName );
		if( pAttr!=NULL ){
			memcpy( usrname, pAttr->lpValue,
				41>pAttr->cbDataLength?pAttr->cbDataLength:41 );
			usrname[40] = '\0';
		}else{
			PL.DP("** Accounting: Missing User Name Attribute");
			return ERROR_INVALID_PARAMETER;
		}

		// Obtain Accounting Status Type
		pAttr = RadiusFindFirstAttribute( pInAttrs, ratAcctStatusType );
		if( pAttr!=NULL ){
			AcctStatusType = pAttr->dwValue;
			PL.DP("** Accounting: Accounting Status Type[%u]", AcctStatusType);
			if( AcctStatusType==1 ){
				PL.DP("** Accounting: User Name[%s]", usrname);
				PL.DP("** Accounting: Vendor ID[%u]",VID);
				PL.DP("** Accounting: NAS Identifier[%s]", NASIdentifier);
				PL.DP("** Accounting: File DSN[%s]", CONNSTR);
				PL.DP("** Accounting: Table name[%s]", TABLENAME);
				PL.DP("** Accounting: Log Table name[%s]", LOGTABLENAME);
				// Obtain Calling Station ID - User's IP or MAC
				ZeroMemory( CallingStationID, 16 );
				pAttr = RadiusFindFirstAttribute( pInAttrs, ratCallingStationId );
				if( pAttr!=NULL ){
					memcpy( CallingStationID, pAttr->lpValue,
						16>pAttr->cbDataLength?pAttr->cbDataLength:16 );
					CallingStationID[16-1] = '\0';
					PL.DP("** Accounting: Calling Station ID[%s]", CallingStationID);
				}else{
					PL.DP("** Accounting: Missing Calling Station ID Attribute");
					return ERROR_INVALID_PARAMETER;
				}
				isql.connect( CONNSTR );
				if( isql.connected )
				{
					// Write DB Login
					IASLoginWriteDB( isql, TABLENAME, usrname, CallingStationID );
					isql.disconnect();
				}else{
					PL.DP( "** Accounting: Cannot connect database" );
					return ERROR_INVALID_PARAMETER;
				}
			}
			if( AcctStatusType==2 ){
				PL.DP("** Accounting: User Name[%s]", usrname);
				PL.DP("** Accounting: Vendor ID[%u]",VID);
				PL.DP("** Accounting: NAS Identifier[%s]", NASIdentifier);
				PL.DP("** Accounting: File DSN[%s]", CONNSTR);
				PL.DP("** Accounting: Table name[%s]", TABLENAME);
				PL.DP("** Accounting: Log Table name[%s]", LOGTABLENAME);
				// Obtain Calling Station ID - User's IP or MAC
				ZeroMemory( CallingStationID, 16 );
				pAttr = RadiusFindFirstAttribute( pInAttrs, ratCallingStationId );
				if( pAttr!=NULL ){
					memcpy( CallingStationID, pAttr->lpValue,
						16>pAttr->cbDataLength?pAttr->cbDataLength:16 );
					CallingStationID[16-1] = '\0';
					PL.DP("** Accounting: Calling Station ID[%s]", CallingStationID);
				}else{
					PL.DP("** Accounting: Missing Calling Station ID Attribute");
					return ERROR_INVALID_PARAMETER;
				}

				// Obtain BandWidth Usage
				pAttr = RadiusFindFirstAttribute( pInAttrs, ratVendorSpecific );
				DWORD count = 2;
				while( pAttr!=NULL ){
					FSVendor = (RADIUS_VSA_FORMAT*)pAttr->lpValue;
					DWORD FSVendorID = ntohl(*(DWORD*)FSVendor->VendorId);
					PL.DP("** Accounting: RADIUS Vendor Id[%u]", FSVendorID);
					if( FSVendorID==VID ){
						// 20080428 - Add quota sharing and accumulate counters check
						if( FSVendor->VendorType==FS_AUTH_BWUSAGE ){
							memcpy( (BYTE*)&bwusage, (BYTE*)FSVendor->AttributeSpecific,
								4 );
							bwusage = ntohl(bwusage);
							PL.DP("** Accounting: D/L BandWidth Usage[%u]", bwusage);
						}else if( FSVendor->VendorType==FS_AUTH_BWUSAGEUL ){
							memcpy( (BYTE*)&bwusageUL, (BYTE*)FSVendor->AttributeSpecific,
								4 );
							bwusageUL = ntohl(bwusageUL);
							PL.DP("** Accounting: U/L BandWidth Usage[%u]", bwusageUL);
						} else if (FSVendor->VendorType == FS_AUTH_BWUSAGESHARE) {
							memcpy((BYTE*)&bwusageSha, (BYTE*)FSVendor->AttributeSpecific, 4);
							bwusageSha = ntohl(bwusageSha);
							PL.DP("** Accounting: Total BandWidth Usage[%u]", bwusageSha);
						}else if( FSVendor->VendorType==FS_AUTH_PRINTQUOTA ){
							memcpy( (BYTE*)&printQuota, (BYTE*)FSVendor->AttributeSpecific,
								4 );
							printQuota = ntohl(printQuota);
							PL.DP("** Accounting: U/L Print Quota[%u]", printQuota);
						}else if( FSVendor->VendorType==FS_AUTH_LOGINMAC ){ // get Bind Mac Address
							int sz =0;
							sz = pAttr->cbDataLength;
							memcpy( &logmac, (BYTE*)FSVendor->AttributeSpecific, sz -6);
							PL.DP("** Accounting: Bind Mac Address [%s], length [%d]", logmac, sz);
						} else if (FSVendor->VendorType == FS_AUTH_ACCUMBW) {
							memcpy((BYTE *)&accumbw, (BYTE *)FSVendor->AttributeSpecific, 4);
							accumbw = ntohl(accumbw);
							PL.DP("** Accounting: Download Accumulate Bandwidth Usage [%u]", accumbw);
						} else if (FSVendor->VendorType == FS_AUTH_ACCUMBWUL) {
							memcpy((BYTE *)&accumbwUL, (BYTE *)FSVendor->AttributeSpecific, 4);
							accumbwUL = ntohl(accumbwUL);
							PL.DP("** Accounting: Upload Accumulate Bandwidth Usage [%u]", accumbwUL);
						} else if (FSVendor->VendorType == FS_AUTH_ACCUMBWSHARE) {
							memcpy((BYTE *)&accumbwSha, (BYTE *)FSVendor->AttributeSpecific, 4);
							accumbwSha = ntohl(accumbwSha);
							PL.DP("** Accounting: Total Accumulate Bandwidth Usage [%u]", accumbwSha);
						} else if( FSVendor->VendorType==FS_AUTH_IBMSG_GLO ){ // get Global in-browser message
							int sz =0;
							sz = pAttr->cbDataLength;
							memcpy( &iBMsg_glo, (BYTE*)FSVendor->AttributeSpecific, sz -6);
							PL.DP("** Accounting: Global in-browser message [%s], length [%d]", iBMsg_glo, sz);
						} else if( FSVendor->VendorType==FS_AUTH_IBMSG_PER ){ // get Personal in-browser message
							int sz =0;
							sz = pAttr->cbDataLength;
							memcpy( &iBMsg_per, (BYTE*)FSVendor->AttributeSpecific, sz -6);
							PL.DP("** Accounting: Personal in-browser message [%s], length [%d]", iBMsg_per, sz);
						}
						else
							PL.DP("** Accounting: Unknown Attribute [%d] " , FSVendor->VendorType);
					}else
						PL.DP("** Accounting: Vendor ID doesn't match");
					pAttr = RadiusFindAttributeByNum( pInAttrs, ratVendorSpecific,
						count );
					count++;
				}
				isql.connect( CONNSTR );
				if( isql.connected )
				{
					// Write DB Login
					IASLogoutWriteDB( isql, TABLENAME, LOGTABLENAME, usrname,
						CallingStationID, bwusage, bwusageUL, logmac, bwusageSha, accumbw, accumbwUL, accumbwSha, iBMsg_glo, iBMsg_per);
					isql.disconnect();
				}else{
					PL.DP( "** Accounting: Cannot connect database" );
					return ERROR_INVALID_PARAMETER;
				}
			}
			if( AcctStatusType==15 ){
				PL.DP("** Accounting: User Name[%s]", usrname);
				PL.DP("** Accounting: Vendor ID[%u]",VID);
				PL.DP("** Accounting: NAS Identifier[%s]", NASIdentifier);
				PL.DP("** Accounting: File DSN[%s]", CONNSTR);
				PL.DP("** Accounting: Table name[%s]", TABLENAME);
				PL.DP("** Accounting: Log Table name[%s]", LOGTABLENAME);
				// Obtain Calling Station ID - User's IP or MAC
				ZeroMemory( CallingStationID, 16 );
				pAttr = RadiusFindFirstAttribute( pInAttrs, ratCallingStationId );
				if( pAttr!=NULL ){
					memcpy( CallingStationID, pAttr->lpValue,
						16>pAttr->cbDataLength?pAttr->cbDataLength:16 );
					CallingStationID[16-1] = '\0';
					PL.DP("** Accounting: Calling Station ID[%s]", CallingStationID);
				}else{
					PL.DP("** Accounting: Missing Calling Station ID Attribute");
					return ERROR_INVALID_PARAMETER;
				}
				// Obtain BandWidth Usage
				pAttr = RadiusFindFirstAttribute( pInAttrs, ratVendorSpecific );
				DWORD count = 1;
				while( pAttr!=NULL ){
					FSVendor = (RADIUS_VSA_FORMAT*)pAttr->lpValue;
					DWORD FSVendorID = ntohl(*(DWORD*)FSVendor->VendorId);
					PL.DP("** Accounting: RADIUS Vendor Id[%u]", FSVendorID);
					if( FSVendorID==VID ){
						if( FSVendor->VendorType==FS_AUTH_PRINTQUOTA ){
							memcpy( (BYTE*)&printQuota, (BYTE*)FSVendor->AttributeSpecific, 4 );
							printQuota = ntohl(printQuota);
							PL.DP("** Accounting: Print Quota[%u]", printQuota);
						}else
							PL.DP("** Accounting: Unknown Attribute");
					}else
						PL.DP("** Accounting: Vendor ID doesn't match");
					pAttr = RadiusFindAttributeByNum( pInAttrs, ratVendorSpecific,
						count );
					count++;
				}
			}
		}
	}
	pAttr = NULL;

	return NO_ERROR;
}

DWORD
WINAPI
RadiusExtensionProcess2(
   PRADIUS_EXTENSION_CONTROL_BLOCK pECB
   )
{
	DWORD result = NO_ERROR;

	PL.DP("** RadiusExtensionProcess2=>");
	//Authentication Packet Accepted
	if( pECB->rcRequestType==1 && pECB->rcResponseType==2 ){
		PL.DP("** RadiusExtensionProcess2(AuthAccept) **");
		result = AuthAccept(pECB);
	}else if( pECB->rcRequestType==4 && pECB->rcResponseType==0 ){
		PL.DP("** RadiusExtensionProcess2(Accounting) **");
		result = Accounting(pECB);
	}
	PL.DP("** <=RadiusExtensionProcess2");
	return result;
}
